//////////////////////////////////////////////////////////////////////////////
//
// $Id: xml.C,v 1.2 2001/06/24 13:49:16 assman Exp $
//
// $Log: xml.C,v $
// Revision 1.2  2001/06/24 13:49:16  assman
// added cvs tags
//
//
//////////////////////////////////////////////////////////////////////////////

#include <xml>

//
// Token tags
//

const string XML::TOKEN::quote = "quote";           // " '
const string XML::TOKEN::equal = "equal";           // =
const string XML::TOKEN::tquestiono = "tquestiono"; // <?
const string XML::TOKEN::tquestionc = "tquestionc"; // ?> 
const string XML::TOKEN::tago = "tago";             // <
const string XML::TOKEN::tagc = "tagc";             // >
const string XML::TOKEN::tbang = "tbang";           // <!
const string XML::TOKEN::tcloseo = "tcloseo";       // </    
const string XML::TOKEN::tempty = "tempty";         // />
//const string XML::TOKEN::tcommento = "tcommento"; // <!--
//const string XML::TOKEN::tcommentc = "tcommentc"; // -->
const string XML::TOKEN::literal = "literal";       // literal string 
const string XML::TOKEN::tcdata = "tcdata";         // cdata tag


const string XML::ERROR::errlist[] = { 
  "success",
    "eof file detected",
    "valid id expected, but not found",
    "bad CDATA tag",
    "bad prolog tag",
    "space elements are not allowed here",
    "bad directive tag found",
    "bad syntax on comment tag",
    "bad syntax on xml tag"
    };

//
// status & error codes
//
	
const int XML::ERROR::success = 0;
const int XML::ERROR::eof = 1;
const int XML::ERROR::badid = 2;
const int XML::ERROR::badcdata = 3;
const int XML::ERROR::badprolog = 4;
const int XML::ERROR::badspaced = 5;
const int XML::ERROR::baddirective = 6;
const int XML::ERROR::badcomment = 7;
const int XML::ERROR::badtag = 8;


//////////////////////////////////////////////////////////////////////////////
//
// Parse_Prolog
//
//////////////////////////////////////////////////////////////////////////////

const int XML::Parser::ESTRICT = 0;
const int XML::Parser::RELAXED = 1;

void XML::Parser::_parse_prolog(istream &in) {
  Token t;
  Token k;
  Token g;

  //
  // get the <?xml entry
  //

  t = _gettoken(in);

  if (!t) {
    _errnum = ERROR::eof;
    return;
  }

  if ((t != TOKEN::tquestiono) || (t.value() != "xml")) {
    _errnum = ERROR::badprolog;
    return;
  }

  g = _gettoken(in);    
  while (g != TOKEN::tquestionc) {
    if (!g) {
      _errnum = ERROR::badprolog;
      return;
    }
      
    //
    // get the equal and the double quotes, if its a literal
    //

    if (g == TOKEN::literal) {
      k = _gettoken(in);
      if ((!k) || (k != TOKEN::equal)) {
	_errnum = ERROR::badprolog;
	return;
      }
      k = _gettoken(in);
      if ((!k) || (k != TOKEN::quote)) {
	_errnum = ERROR::badprolog;
	return;
      }
      //
      // attribute contents:
      // name of the directive: t.value();
      // name of the attr:      g.value();
      // contents of the attr:  k.value();
      //
      //cout << "Found attr: " << g.value() << " with: " << k.value() << endl;
    } 
    else {
      if (g != TOKEN::tquestionc) {
	_errnum = ERROR::badprolog;
	return;
      }
      //
      // its ok
      //
      break;
    }
    g = _gettoken(in);
  }

  return;
}

//////////////////////////////////////////////////////////////////////////////
//
// Parse_Directives
//
//////////////////////////////////////////////////////////////////////////////

void XML::Parser::_parse_directives(istream &in) {
  Token t;
  Token k;

  t = _gettoken(in);
  while (t != TOKEN::tago) {

    if (!t) {
      _errnum = ERROR::baddirective;
      return;
    }

    if (t != TOKEN::tbang) {
      _errnum = ERROR::baddirective;
      return;
    }

    //
    // skip till the end of the tags
    // <! case
    //

    if (t == TOKEN::tbang) {
      k = _gettoken(in);
      while (k != TOKEN::tagc) {
	if (!k) {
	  _errnum = ERROR::baddirective;
	  return;
	}
	k = _gettoken(in);
      }
    }

    t = _gettoken(in);
  }
  //
  // save token for later use
  //
  _ungettoken(t);
}


//////////////////////////////////////////////////////////////////////////////
//
// Parse_Tag
//
//////////////////////////////////////////////////////////////////////////////


XML::Tree XML::Parser::_parse_tag(istream &in) {
  Token t;
  Token g;
  Token k;
  Token l;
    
  Tree l_tree; // subtrees local one
  Tree g_tree; // generated by recursive call
  Tree d_tree; // dummy tree

  TagEntry tag;     // a tag

  t = _gettoken(in);
  if (!t) {
    _errnum = ERROR::eof;
    return(d_tree);
  }

  if (t != TOKEN::tago) {
    _errnum = ERROR::badtag;
    return(d_tree);
  }

  tag.name(t.value());

  g = _gettoken(in);
  while ( (g != TOKEN::tagc) && (g != TOKEN::tempty) ) {

    if (!g) {
      _errnum = ERROR::badtag;
      return(d_tree);
    }

    //
    // get the equal and the tag environment
    //

    if (g == TOKEN::literal) {
      k = _gettoken(in);
      if ((!k) || (k != TOKEN::equal)) {
	_errnum = ERROR::badtag;
	return(d_tree);
      }
      k = _gettoken(in);
      if ((!k) || (k != TOKEN::quote)) {
	_errnum = ERROR::badtag;
	return(d_tree);
      }
      //
      // attribute contents:
      // name of the directive: t.value();
      // name of the attr:      g.value();
      // contents of the attr:  k.value();
      //
      //cout << "Found attr: " << g.value() << " with: " << k.value() << endl;
      tag.attr.add(g.value(),k.value());
    }
    else {
      if ( (g != TOKEN::tagc) && (g != TOKEN::tempty) ) {
	_errnum = ERROR::badtag;
	return(d_tree);
      }
    }
    g = _gettoken(in);
  }
    
  //
  // this is an empty tag
  // so build the tree and return it
  //

  if (g == TOKEN::tempty) {
      //cout << "get a full emtpy tag" << endl;
    l_tree.add(tag);
  }

  //
  // this is a compound tag, so we have 
  // to get its content, and wait for the
  // end-of-tag
  //

  if (g == TOKEN::tagc) {
      //cout << "get a compound tag" << endl;
    l_tree.add(tag);
      
    //
    // here, we can have:
    //  - comments (stripped by the lexical layer)
    //  - more xml tags 
    //  - text data
    //
    // use raw mode to get next token
    //

    l = _gettoken(in,1); 

    while ( l != TOKEN::tcloseo ) {

      if (!l) {
	_errnum = ERROR::eof;
	return(d_tree);
      }

      //
      // check for error tags
      //

      if ( (l != TOKEN::literal) && (l != TOKEN::tago) && 
	   (l != TOKEN::tcloseo) && (l != TOKEN::tcdata) ) {
	_errnum = ERROR::badtag;
	return(d_tree);
      }

      //
      // process literal contents here
      // for now, do nothing.
      //

      if ( (l == TOKEN::literal) || (l == TOKEN::tcdata) ) {
	  //if (l == TOKEN::tcdata) {
	    //cout << l.value() << endl;
	  //}
	tag.content(l.value());
	//
	// overwrite the contents with this new ones
	//
	l_tree.add(tag);             
      }

      //
      // if found a tag open, call recursivelly to this function
      // to process it.
      //

      if (l == TOKEN::tago) {
	_ungettoken(l);
	g_tree = _parse_tag(in);
	if (g_tree.active()) l_tree.addchild(g_tree);
	if (_errnum) return(d_tree);
      }
	
      l = _gettoken(in);
    }
      


    //
    // if we fall here, it is due we have a close tag
    //
      
    if (l == TOKEN::tcloseo) {
      //
      // if tag close, test if match the initial label of the tag
      // if not equal set and error
      // else, go out
      //
      if (l.value() != t.value()) {
	_errnum = ERROR::badtag;
	return(d_tree);
      }
      //
      // finally, read the close tag
      // and build the tree
      //

      l = _gettoken(in);
      if (!l) {
	_errnum = ERROR::eof;
	return(d_tree);
      }
      if (l != TOKEN::tagc) {
	_errnum = ERROR::badtag;
	return(d_tree);
      }
    }

  } // end of compound token

  // 
  // return a copy of the builded tree
  //
  return(l_tree);
}

//////////////////////////////////////////////////////////////////////////////
//
// Read_Id
//
//////////////////////////////////////////////////////////////////////////////


XML::Parser::Token XML::Parser::_read_id(istream &in, string toktype) {
  char ch;
  string lit;
  Token e;
  Token t;

  in >> ch;
    
  if (!in.eof()) {
    if ((ch != '-') && (ch != ':') && (!isalpha(ch))) {
      _errnum = ERROR::badid;
      return(e);
    }
    lit = ch;
    in >> ch;
    while (!in.eof()) {
      if ((ch == '-') || (ch == ':') || (ch == '_') ||
	  (ch == '.') || (isalnum(ch))) {
	lit += ch;
      }
      else {
	in.unget();
	t.init(toktype,lit);
	return(t);
      }
      in >> ch;
    }
    if (in.eof()) {
      _errnum = ERROR::eof;
      return(e);
    }
  }
  //
  // if we fall here, its an error
  //
  _errnum = ERROR::eof;
  return(e);
}

//////////////////////////////////////////////////////////////////////////////
//
// Read_Cdata
//
//////////////////////////////////////////////////////////////////////////////

XML::Parser::Token XML::Parser::_read_cdata(istream &in, string toktype) {
  Token e;
  Token t;
  char ch,ch2,ch3;
    
  string lit="";
  string cdata_o = "[CDATA[";
  unsigned int n=0;
  in >> ch;
  if (in.eof()) {
    _errnum = ERROR::badcdata;
    return(e);
  }
  if (ch == '[') {
    while (!in.eof()) {
      if (ch == cdata_o[n]) {
	n++;
	if (n == cdata_o.length()) {
	  //
	  // read characters until end found
	  // <![CDATA[ accepted
	  //
	  in >> ch;
	  while (!in.eof()) {
	      
	    if ( ch == ']') {
	      in >> ch2;
	      if (in.eof()) {
		_errnum = ERROR::badcdata;
		return(e);
	      }
	      if (ch2 == ']') {
		in >> ch3;
		if (in.eof()) {
		  _errnum = ERROR::badcdata;
		  return(e);
		}
		if (ch3 == '>') {
		  //
		  // end of cdata section reached,
		  // return the literal item
		  //
		  t.init(toktype,lit);
		  return(t);
		}
		in.unget();
	      } 
	      in.unget();
	    } 
	    lit += ch;
	    in >> ch; 
	  }
	  if (in.eof()) {
	    _errnum = ERROR::badcdata;
	    return(e);
	  }
	}
      }
      else {
	_errnum = ERROR::badcdata;
	return(e);
      }
      in >> ch;
    }
    if (in.eof()) {
      _errnum = ERROR::badcdata;
      return(e);
    }
  }
  else {
    in.unget();
    return(e);
  }
}

//////////////////////////////////////////////////////////////////////////////
//
// Get_Protected_Env
//
//////////////////////////////////////////////////////////////////////////////

XML::Parser::Token 
XML::Parser::_get_protected_env(istream &in, string toktype, char env) {
  string lit = "";
  char ch2, ch3;
  Token t;
  Token e;

  in >> ch2;
  while (ch2 != env) {
    if (in.eof()) {
      _errnum = ERROR::eof;
      return(e);
    }
    if (ch2 == '\\') {
      in >> ch3;
      if (in.eof()) {
	_errnum = ERROR::eof;
	return(e);
      }
      if (ch3 == env) {
	lit += env;
      }
      else {
	  lit += "\\";
	  lit += ch3;
      }
    }
    else {
      lit += ch2;
    }
    in >> ch2;
  }
  t.init(toktype,lit);
  return(t);
}

//////////////////////////////////////////////////////////////////////////////
//
// Strip_Comment
//
//////////////////////////////////////////////////////////////////////////////


void XML::Parser::_strip_comment(istream &in) {
  char ch;

  in >> ch;
    
  while (!in.eof()) {
    if (in.eof()) { 
      _errnum = ERROR::eof;
      return;
    }

    if (isspace(ch)) {
      if (ch == '\n') _currline++;

      in >> ch;
      if (in.eof()) { 
	_errnum = ERROR::eof;
	return;
      }
      if (ch == '-') {
	in >> ch;
	if (in.eof()) { 
	  _errnum = ERROR::eof;
	  return;
	}
	if (ch == '-') {
	  in >> ch;
	  if (in.eof()) {
	    _errnum = ERROR::eof;
	    return;
	  }
	  if (ch == '>') {
	    //
	    // comment fully sttriped
	    // so return
	    //
	    return;
	  }
	  in.unget();
	}
	in.unget();
      }
      in.unget();
    }
    in >> ch;
  }
}

//////////////////////////////////////////////////////////////////////////////
//
// Gettoken
//
//////////////////////////////////////////////////////////////////////////////


XML::Parser::Token XML::Parser::_gettoken(istream &in, int mode=0) {
  Token t;
  Token e;
  Token k;

  string tmptok="";
  string lit="";
  char ch,ch1,ch2,ch3,ch4;

  _errnum = ERROR::success; // no error 

  if (_tok_stack.size()==0) {

    in.unsetf(ios::skipws);  // no skip whitespaces
    while (!in.eof()) {
	
      in >> ch;

      // blank spaces ///////////////////////////////

      if (isspace(ch)) {                       
	if (ch == '\n') _currline++;

	if (tmptok != "") {                   
	  in.unget();
	  t.init(TOKEN::literal,tmptok);
	  return(t);
	}
	else continue;
      }

      // equal //////////////////////////////////////

      if (ch == '=') { 
	in >> ch2;
	if (in.eof()) {
	  _errnum = ERROR::eof;
	  return(e);
	}
	if (isspace(ch2)) {
	  _errnum = ERROR::badspaced;
	  return(e);
	}
	in.unget();
	t.init(TOKEN::equal,"");
	break;
      }
	 
      // close tag //////////////////////////////////

      if (ch == '>') { 
	t.init(TOKEN::tagc,"");
	break;
      }

      // double quote      //////////////////////////
      // and single quote ///////////////////////////
 
      if (ch == '\"') {
	t = _get_protected_env(in,TOKEN::quote,'\"');
	if (!t) return(e);
	break;
      }

      if (ch == '\'') {
	t = _get_protected_env(in,TOKEN::quote,'\'');
	if (!t) return(e);
	break;
      }

      // empty token ////////////////////////////////

      if (ch == '/') {
	in >> ch1;
	if (in.eof()) { 
	  _errnum = ERROR::eof; 
	  return(e); 
	}
	if (ch1 == '>') {
	  t.init(TOKEN::tempty,""); 
	  break;
	}
	else { 
	  in.unget();
	  in.unget();
	}
      }

      // open token /////////////////////////////////

      if (ch == '<') {
	in >> ch1;
	if (in.eof()) {
	  _errnum = ERROR::eof;
	  return(e);
	}
	  
	if (ch1 == '?') {
	  t = _read_id(in,TOKEN::tquestiono);
	  if (!t) return(e);
	  break;
	}

	// close token /////////////////////////////

	if (ch1 == '/') { 
	  t = _read_id(in,TOKEN::tcloseo);
	  if (!t) return(e);
	  break;
	}

	// bang token //////////////////////////////

	if (ch1 == '!') {
	  in >> ch2;
	  if (in.eof()) {
	    _errnum = ERROR::eof;
	    return(e);
	  }
	  if (ch2=='-') {
	    in >> ch3;
	    if (in.eof()) {
	      _errnum = ERROR::eof;
	      return(e);
	    }

				// comment open ////////////////////////////
				// strip here the comments /////////////////
	      
	    if (ch3=='-') {

	      in >> ch4;
	      if (in.eof()) {
		_errnum = ERROR::eof;
		return(e);
	      }

	      if (!isspace(ch4)) {
		_errnum = ERROR::badcomment;
		return(e);
	      }
		
	      _strip_comment(in);
	      if (_errnum) {
		_errnum = ERROR::badcomment;
		return(e);
	      }
	      in >> ch;
	      continue;
	    }
	    else in.unget();
	  }
	  else in.unget();

	  // tag bang ////////////////////////////////

	  // parse <![CDATA[ ..... ]]> as "protected" environment
	    
	  t = _read_cdata(in,TOKEN::tcdata);

	  if (!t) {
	    if (_errnum) return(e);
	    else {
	      t = _read_id(in,TOKEN::tbang);
	      if (!t) return(e);
	      break;
	    }
	  }
	  else  break;
	}
	else in.unget();

	// tag open ////////////////////////////////

	t = _read_id(in,TOKEN::tago);
	if (!t) return(e);
	break;
      }


      // question close //////////////////////////

      if (ch == '?') {
	in >> ch1;
	if (in.eof()) {
	  _errnum = ERROR::eof;
	  return(e);
	}
	if (ch1 == '>') {
	  t.init(TOKEN::tquestionc,"");
	  break;
	    
	} else in.unget();
      }

      //
      // read the identifiers
      // and the contents using the way
      // there must be NO spaces between the id, the equal and the quote
      // characters so handle it here
      //

      //
      // default mode
      // assume ids are properties
      //
      if (mode == 0) {

	if (!t) {
	  in.unget();
	  k = _read_id(in,TOKEN::literal);
	  tmptok = k.value();
	  in >> ch;
	  if (in.eof()) {
	    _errnum = ERROR::eof;
	    return(e);
	  }
	  if (isspace(ch)) {
	    _errnum = ERROR::badspaced;
	    return(e);
	  }
	  in.unget();
	}

      }
      else {
	//
	// assume it is text contents, so don't stop till
	// eof || a special tag was found (we look for <)
	//

	in.unget();
	in >> ch;
	while (ch != '<') {
	  if (in.eof()) {
	    _errnum = ERROR::eof;
	    return(e);
	  }
	  //
	  // translate all the space elements into spaces
	  //
	  if (isspace(ch)) {
	    ch = ' ';
	    if (ch == '\n') _currline++;
	  }
	  tmptok +=ch;
	  in >> ch;
	}
	in.unget();
      }
    }
    // while

    if (in.eof()) {
      _errnum = ERROR::eof;
    }
      
    if (tmptok != "")  {
      _tok_stack.push(t);
      e.init(TOKEN::literal,tmptok);
      return(e);
    }
    else return(t);

  }
     
    
  t = _tok_stack.top();
  _tok_stack.pop();
    
  return(t);
}
