//////////////////////////////////////////////////////////////////////////////
//
// $Id: xml.h,v 1.3 2001/07/21 14:33:28 assman Exp $
//
// Quick implentation of a conforming xml document parser
// not a fully implementation of a 1.0 xml conformant parser,
// but it does the trick
// 
//////////////////////////////////////////////////////////////////////////////
//
// <! .... > is ignored
// <? .... ?> parsed only for xml field (prolog)
// <!-- ... --> ignored also (they are comments)
// 
// protected enviroments are on the way
// parse <![DATA[ ..... ]]> as "protected" environment
// also protect data between "" and ''
// allows scaped quoted characters (\" and \')
// note that spaces are not allowed between = ids and quotes
// note that between comments we need to have a black space
// comments are removed at syntax level
//
// LOG
//
// 1 First version of the tokenizer 28/apr/2001 (sat 13:42)
//   I have the parser working for 1 tag at 28/apr/2001 (sat 20:00)
//
// 2 At 29/apr/2001 (sun 11:15) I have the xml parser working with all
//   type of tags and nested elements. In literal items (not cdata) all
//   the space characters are translated into spaces
//
// 3 At 29/apr/2001 (sun 16:32) the STL tree class has been implemented
//   and it is working. I have included the tree class into the parser,
//   and now the nodes are stored into the tree. I have to write down a 
//   property container, so we can insert the attributes of each tag 
//
// 4 At 30/apr/2001 (mon 09:51) changed all the input parameters to the
//   tree class into references, so the parameter passing to the method
//   is speeded up
//
// 5 At 30/apr/2001 (mon 13:38) TagEntry has been included and the parser
//   is now fully functional. We are in beta testing for this product.
//
// 6 At 15/may/2001 (tue 16:47) xmlparser has been splitted into various
//   files, and now it builds its own library.
//
// 7 At 15/may/2001 (tue 16:57) a line count has been added into the error
//   support class, so all the xml messages have the line when the error
//   happens. Some hack into get_parser has been required, basically, add
//   one when a \n is found. Also get_cdata an get_comment has been touched
//
//////////////////////////////////////////////////////////////////////////////
//
// CVS LOG
//
// $Log: xml.h,v $
// Revision 1.3  2001/07/21 14:33:28  assman
// Fixed some bugs
//
// Revision 1.2  2001/06/24 13:49:16  assman
// added cvs tags
//
//
//////////////////////////////////////////////////////////////////////////////

#ifndef __XML_H__
#define __XML_H__

#include <cstdio>
#include <cstdlib>
#include <cctype>
#include <strstream>
#include <fstream>

#include <iostream>

#include <stack>
#include <string>

#include <errormgmt>
#include <paircontainer>
#include <tree>
#include <numconv>
#include <regexpressions>

//////////////////////////////////////////////////////////////////////////////
//
// The XML namespace & components
//
//////////////////////////////////////////////////////////////////////////////

namespace XML {

    namespace TOKEN {

	//
	// Token tags
	//

	const extern string quote;             // " '
	const extern string equal;             // =
	const extern string tquestiono;        // <?
	const extern string tquestionc;        // ?> 
	const extern string tago;              // <
	const extern string tagc;              // >
	const extern string tbang;             // <!
	const extern string tcloseo;           // </    
	const extern string tempty;            // />
	//const extern string tcommento;       // <!--
	//const extern string tcommentc;       // -->
	const extern string literal;           // literal string 
	const extern string tcdata;            // cdata tag

    }

    namespace ERROR {
	const extern string errlist[];

	//
	// status & error codes
	//
	
	const extern int success;
	const extern int eof;
	const extern int badid;
	const extern int badcdata;
	const extern int badprolog;
	const extern int badspaced;
	const extern int baddirective;
	const extern int badcomment;
	const extern int badtag;
	
    }

    typedef pair_container<string,string> AttrContainer;
  
    //
    // the entry for the tree
    //

    class TagEntry {
    public:  
	TagEntry() { init(); }

	TagEntry(string name, string content="") {
	    init();
	    _name = name;
	    _content = content;
	}

	~TagEntry() { init(); }

	TagEntry(const TagEntry &e) {
	    _name = e._name;
	    _content = e._content;
	    attr = e.attr;
	}
  
	void init(void) {
	    _name = "";
	    _content = "";
	    attr.init();
	}
  
	//  void addprop(string attr, string value) {
	//  _attr.add(attr,value);
	//  }

	// void addprop(AttrContainer at) {
	//  int n;
	//  for (n=0; n< at.size(); n++) {
	//    _attr.add(at.key(n),at[n]);
	//  }
	//}

	string name(void) { return(_name); }
	void name(string name) { _name = name; }

	string content(void) { return(_content); }
	void content(string content) { _content = content; }

	//
	// to allow equity test in tree searches
	// is equal if the name is equal
	//
	
	bool operator==(const TagEntry &t) {
	  return(_name == t._name ? true : false);
	}

	bool operator==(const string s) {
	  return(_name == s ? true : false);
	}
	
	bool operator!=(const string s) {
	  return(_name == s ? false : true);
	}


	//
	// show the attr interface
	//

	AttrContainer attr;

    protected:

	string        _name;
	string        _content;


	friend ostream &operator<<(ostream &o, TagEntry &t) {
	    int n;
	    o << t.name() << "<" << t.content() << ">";
	    for (n=0; n< t.attr.size(); n++) {
		o << " " << t.attr.key(n) << ": " << t.attr[n];
	    }
	    o << endl;
	    return(o);
	}


    };


    //  *WARNING*WARNING*WARNING**WARNING*WARNING*WARNING*   
    //
    // this is a copy of the tree template include in tree.h
    // the problem is that I can't inherit from the template
    // and overload a function. This should be fixed so we
    // mantain only one copy of the tree class in the system
    //
    //  *WARNING*WARNING*WARNING**WARNING*WARNING*WARNING*


    class Tree {
    public:
      
      Tree() { init(); }
      
      Tree(const TagEntry &data) { add(data); }
      
      Tree(const Tree &t) {
	init();
	_active = t._active;
	_data   = t._data;
	_childs = t._childs;
      }
      
      virtual ~Tree() {
	init();
      }

      void init(void) {
	_active = false;
	if (_childs.size() > 0) {
	  _childs.erase(_childs.begin(),_childs.end());
	}
      }
      
      bool active(void) {
	return(_active);
      }
      
      Tree search(TagEntry data) {
	unsigned int n;
	Tree tmp_a;
	if (_active) {
	  if (data == _data) {
	    Tree tmp(*this);
	    return(tmp);
	  }
	  else {
	    for (n=0; n< _childs.size(); n++) {
	      tmp_a = _childs[n].search(data);
	      if (tmp_a.active()) return(tmp_a);
	    }
	  }
	}
	return(tmp_a);
      }

      //
      // the goal is to test if the ttree regexp matches
      // the current tree. To archieve this, we need to
      // check each attribute, at the same level
      //

      bool match(Tree &ttree) {
	  Regexpression reg;
	  int i,j;
	  string tmp_i;
	  string tmp_j;
	  bool found;

	  //
	  // test if the current tree is active and if 
	  // the contents of the node maches the regexp
	  // tree node
	  //

	  if (!_active) 
	      return(false);
	  
	  if (_data.name() != ttree.data().name())
	      return(false);
	  
	  if ((ttree.data().content() != "") && 
	      (!reg.regmatch(ttree.data().content(),_data.content()))) 
	      return(false);

	  //
	  // test all the attributes in the current
	  // node.
	  //

	  for (i = 0; i < ttree.data().attr.size(); i++) {

	      if (ttree.data().attr[i] == "")  // if empty, continue
		  continue;

	      tmp_i = ttree.data().attr.key(i);
	      tmp_j = _data.attr[tmp_i];

	      // not found

	      if ((tmp_j == "") || (!reg.regmatch(ttree.data().attr[i],tmp_j)))
		  return(false);

	  }       
	     
	  //
	  // first, process all the childs in the two trees. Note
	  // that the childs may be unsorted, so we have to find
	  // them and launch the search for every subtree
	  //

	  for (i = 0; i < ttree.size(); i++) {
	      for (j = 0; j < (signed int)_childs.size(); j++) {
		  if (ttree[i].name() == _childs[j].data().name()) {
		    Tree t;
		    t = ttree[i];
		    if (!_childs[j].match(t))
		      return(false);
		    
		    found = true;
		    break;
		  }
		  else 
		    found = false;
	      }
	      //
	      // if the matched child isn't found, return false
	      //

	      if (!found)
		return(false);
	  }

	  //
	  // if we reached this, the tree match
	  //

	  return(true);
      }

      void print(unsigned int indent=0) {
	unsigned int n=0;
	string s_ind = "";
	for (n=0; n< indent; n++) s_ind += " ";
      
	if (_active) {
	  cout << s_ind << "node: " << _data;
	  for (n=0; n < _childs.size(); n++) {
	    _childs[n].print(indent+5);
	  }
	}
      }
  
      int size(void) {                               
	return _childs.size(); 
      }  

      void add(const TagEntry &data) {
	_data = data;
	_active = true;
      }
  
      bool addchild(const Tree &data) {
	if (!_active) return(false);
	_childs.push_back(data);
	return(true);
      }


      TagEntry data(void) {
	return(_data);
      }

      TagEntry *dataref(void) {
	  return(&_data);
      }
  
      // 
      // store a copy of the element
      // instead the element itself
      // note that for destruction pourpouses, the element
      // must have its own destructors defined
      //

      bool addchild(const TagEntry &data) {
	Tree *t;
	if (!_active) return(false);
	t = new Tree(data);
	addchild(*t);
	delete t;
	return(true);
      }

      void removechild(int idx) {
	  _childs.erase(_childs.begin()+idx);
      }

      TagEntry operator[](unsigned int idx) {
	return(_childs[idx]._data);
      }


      TagEntry operator[](string idx) {
	  unsigned int n;
	  TagEntry empty;
	  for (n=0; n < _childs.size(); n++)
	      if (_childs[n].data().name() == idx)
		  return(_childs[n].data());
	  return(empty);
      }


      //
      // serialize the tree into a XML string 
      // this is the core of the program, due
      // we have to serialize lots of events
      //

      string to_xml(void) {
	unsigned int n;
	string cad = "";
	
	if (_active) {
	  cad = "<" + _data.name();

	  for (n=0; (int)n < _data.attr.size(); n++) {
	    cad = cad + " " + _data.attr.key(n) + 
	      "=\"" + _data.attr[n] + "\"";
	  }

	  if ((_data.content() == "") && (_childs.size()==0)) {
	      cad = cad + "/>";
	      return(cad);
	  }

	  cad = cad + ">" + _data.content();
	  
	  for (n=0; n < _childs.size(); n++) {
	    cad = cad + _childs[n].to_xml();
	  }

	  cad = cad + "</" + _data.name() + ">";
	}
	return(cad);
      }

      //
      // add two functions to manage the data in the childs, 
      // accessing them using a string as index
      //

      Tree getchild(int idx) { return(_childs[idx]); }

      string getprop(string child, string name);
      void setprop(string child, string name, string value);


    protected:
      vector<Tree> _childs;   // the childs
      TagEntry     _data;     // the data 
      bool         _active;   // this node has elements
    };

    //
    // a type to manage the tree
    // can't be a class, because the recursive issue
    //

    //typedef tree<TagEntry> ParseTree;
    //typedef Tree ParseTree ;

    //
    // a class to manage the errors
    //
    
    class ParserError: public GenericError {
    public:
	ParserError()  { clearerr(); _currline = 0; }
	int errnum(void) { return(_errnum); }
	string strerr(void) { 
	    string err = "";
	    err = XML::ERROR::errlist[_errnum] + " on line " + 
		TOOLS::itoa(_currline);
	    return(err); 
	}
	void clearerr(void) {
	    _errnum = XML::ERROR::success;
	}
	
    protected:
	int _errnum;
	int _currline;
    };
    

    ////////////////////////////////////////////////////////////////////////
    //
    // the xml parser
    //
    ////////////////////////////////////////////////////////////////////////
    
    class Parser: public ParserError {
    public:

	Parser(): ParserError() {
	    init();
	}

	virtual ~Parser() {
	    init();
	}
	
	//
	// we have to put here a init function to reset the state of the 
	// parser, that is, error is set to success, and all the stack is
	// cleared
	//

	void init(void) {
	    _errnum = ERROR::success;

	    while (_tok_stack.size()>0) {
		_tok_stack.pop();
	    }

	    _tree.init();
	}
	
	static const int ESTRICT;
	static const int RELAXED;

	Tree parse(istream &in, int mode=ESTRICT) {
	    Token t;

	    //
	    // if ESTRICT mode is selected
	    // process prolog & directives
	    // so the xml is a fully conformant xml file, else
	    // only parse the body (useful to use when we are parsing
	    // subdata
	    //

	    if (mode == ESTRICT) {
	      _parse_prolog(in);
	      if (errnum()) return(_tree);

	      _parse_directives(in);
	      if (errnum()) return(_tree);
	    }
	      
	    //
	    // process the body
	    //

	    _parse_xml_body(in);
	    if (errnum()) return(_tree);

	    return(_tree);
	}

    protected:

	//
	// this is a token to do the parsing
	//

	class Token {
	public:

	    Token(): 
		_id(""), 
		_value("") { }

	    Token(string id, string value): 
		_id(id),
		_value(value) { }

	    Token(const Token &t) {
		_id = t._id;
		_value = t._value;
	    }

	    void init(const string id, const string value) {
		_id = id;
		_value = value;  
	    }
    
	    string value(void) { 
		return(_value); 
	    }

	    string id(void) { 
		return(_id); 
	    }

	    bool operator==(const string str) {
		return(_id == str ? true :false);
	    }

	    bool operator!=(const string str) {
		return(!operator==(str));
	    }

	    bool operator==(const bool val) {
		return(_id != "" && val ? true : false);
	    }

	    bool operator!(void) {
		return(_id=="" && _value=="" ? true :false);
	    }

	protected:
	    string _id;
	    string _value;

	};

	//
	// parse prolog
	// <?xml attr_1="..." attr_2="..." ...>

	void _parse_prolog(istream &in);

	//
	// _parse_directives
	// parse all between the prolog and the effective xml body
	//
  
	void _parse_directives(istream &in);

	//
	// _parse_xml_body
	// this is the part when we parse the xml body and build the tree
	//
  
	void _parse_xml_body(istream &in) {
	    _tree = _parse_tag(in);
	    //_tree.print();
	    if (_errnum) return;
	}

	//
	// _parse_tag
	// get a xml tag with its attributes
	// this is the core of the parser
	// XXX 
 
	Tree _parse_tag(istream &in);

	//
	// read the id from the tag. This rule is applied also to the 
	// atributes of a tag
	//
  
	Token _read_id(istream &in, string toktype);


	Token _read_cdata(istream &in, string toktype);  
	Token _get_protected_env(istream &in, string toktype, char env);  
	void _strip_comment(istream &in);  

	//
	// tokenizer -> get & unget tokens
	// throws XML::eof if eof found
	//
  
	Token _gettoken(istream &in, int mode=0);
  
	void _ungettoken(const Token &tok) {
	    _tok_stack.push(tok);
	}

	//
	// variables
	//

	stack<Token> _tok_stack; 
	Tree         _tree;
    };
}

#endif


